import warnings

from cqrs.exceptions import InvalidCommandParameter


RESERVED_ATTRS = ['metadata','handler','accept','serialize','as_dict',
    '_set_parameters', '__getitem__','__iter__','__len__','inspector',
    'as_tuple','__hash__','__eq__','Meta', 'set_result', '_result',
    'get_identity','qualified_name']


class CommandParameter(object):
    empty_values = ['', list(), tuple(), set(), dict(),]
    creation_counter = 0

    def __init__(self, required=True, nullable=False, type=None, empty_values=None,
        default=None, unique=False, primary_key=False):
        self.required = required
        self.type = type or (lambda command, x: x)
        self.nullable = nullable
        if empty_values:
            self.empty_values += list(empty_values)
        CommandParameter.creation_counter += 1
        self.creation_counter = CommandParameter.creation_counter
        self.default = default
        self.unique = False
        self.primary_key = primary_key

    def validate(self, value):
        """
        Validates the input value of a command parameter.
        """

        if value is None and (self.default is not None):
            value = self.default if not callable(self.default)\
                else self.default()

        if value is None and not self.nullable:
            raise InvalidCommandParameter(self.attname,
                "'{0}' is a required parameter".format(self.attname))

        # Assess if the provided value is empty.
        if value is not None:
            is_empty = (hasattr(value, '__len__') and (len(value) == 0))\
                or (value in self.empty_values)
            if is_empty and self.required:
                raise InvalidCommandParameter(self.attname,
                    "'{0}' is a required parameter".format(self.attname))

        return value


    def contribute_to_class(self, cls, name):
        """
        Contributes the :class:`CommandAttribute` to a newly created
        :class:`~cqrs.Command` subclass.

        Args:
            cls: the newly created class.
            name (str): attribute name of the command parameter.
        """
        if name in RESERVED_ATTRS:
            raise ValueError("'{0}' is a reserved name.".format(name))
        self.attname = name
        setattr(cls, name, self)
        cls.__parameters__ += tuple([self])

    def __get__(self, instance, cls):
        return instance._parameters[self.attname] if instance\
            else self


class CommandAttribute(CommandParameter):

    def __init__(self, *args, **kwargs):
        warnings.warn("CommandAttribute is deprecated, use CommandParameter.",
            DeprecationWarning)
        CommandParameter.__init__(self, *args, **kwargs)


import abc


class CommandHandler(metaclass=abc.ABCMeta):
    """
    Base class for command handlers. Subclasses must override
    the :meth:`CommandHandler.handle` method.
    """

    @abc.abstractmethod
    def handle(self, command, *args, **kwargs):
        """
        Execute the command. Receives the command and additional
        parameters passed to the gateway.
        """
        pass

    def run_command(self, command, func, *args, **kwargs):
        # Private method to allow setting up a specific command
        # execution context.
        return func(command, *args, **kwargs)



import abc
import collections
import threading
import random
import sys

class IGateway(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def dispatch(self, command, *args, **kwargs):
        """
        Executes the given command.
        """
        pass

    def put(self, command, *args, **kwargs):
        """Put a command an assess if it can be executed; proceed to
        invoke the command-runner to execute. Returns a :class:`uuid4
        cqrs.Transaction` instance representing the transaction.

        Args:
            command (cqrs.Command): a concrete command implementation.
            **extra_metadata: extra metadata parameters.

        Returns:
            cqrs.Transaction
        """
        self.dispatch(command, *args, **kwargs)


    @abc.abstractmethod
    def dispatch(self, command, *args, **kwargs):
        """Dispatches the command to the command-processing gateway
        backend.

        Args:
            command: the command to dispatch.
        """
        raise NotImplementedError



from cqrs.command.command import Command
from cqrs.command.attribute import CommandAttribute
from cqrs.command.attribute import CommandParameter
from cqrs.command.exceptions import CommandValidationError
from cqrs.command.inspector import CommandInspector
from cqrs.command.meta import CommandMeta


import importlib


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class
    designated by the last name in the path. Raise :exc:`ImportError`
    if the import failed.

    Args:
        dotted_path: a string containing the fully-qualified path
            to an attribute in a Python module.

    Returns:
        object: the attribute specified by `dotted_path`.
    """
    dotted_path = str(dotted_path)
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        msg = "{0} doesn't look like a module path".format(dotted_path)
        raise ImportError(msg)

    module = importlib.import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        msg = 'Module "{0}" does not define a "{1}" attribute/class'.format(
            dotted_path, class_name)
        raise ImportError(msg)



import abc
import collections
import uuid
import datetime

from cqrs.command.inspector import CommandInspector
from cqrs.command.meta import CommandMeta
from cqrs.exceptions import InvalidCommandParameters
from cqrs.exceptions import InvalidCommandParameter
from cqrs.const import PENDING

__all__ = ['Command']


isoformat = lambda x: datetime.datetime.strptime("%Y-%m-%d %H:%M:%S.%f")
datetime_from_string = lambda x: isoformat(x) if (x is not None) else None


class Command(metaclass=CommandMeta):
    """
    Describes a destructive command on a certain part of a system.
    """

    @property
    def metadata(self):
        return self._metadata

    @property
    def inspector(self):
        return self._inspector

    def __init__(self, *args, **kwargs):
        self._inspector = CommandInspector()
        params = {}
        errors = []
        for param in self._inspector.get_parameters(self):
            value = kwargs.get(param.attname)
            try:
                value = param.validate(value)
                params[param.attname] = value
            except InvalidCommandParameter as e:
                errors.append(e)
        if errors:
            raise InvalidCommandParameters(parameters=errors)
        self._set_parameters(params)
        self._result = None

    @classmethod
    def handler(cls, handler_class):
        """
        Class-decorator to use on a :class:`~cqrs.CommandHandler`
        implementation to register it for `cls`.
        """
        handler_class.command_type = cls
        cls.handler = handler_class
        return handler_class

    def as_dict(self):
        """Project the command as a dictionary."""
        return self.inspector.as_dict(self)

    def _set_parameters(self, params):
        self._parameters = {}
        for param in self._inspector.get_parameters(self):
            self._parameters[param.attname] = params.pop(param.attname)

    def __getitem__(self, key):
        if key not in self._attributes:
            raise KeyError(key)
        return getattr(self, key)

    def __iter__(self):
        return self.as_dict().__iter__()

    def __len__(self):
        return len(self.as_dict().keys())

    def __hash__(self):
        return self.inspector.hash(self)

    def __eq__(self, other):
        return hash(self) == hash(other)



import collections

from cqrs.command.attribute import CommandAttribute as CommandParameter


class CommandMeta(type(collections.Mapping)):

    def __new__(cls, name, bases, attrs):
        super_new = super(CommandMeta, cls).__new__

        # six.withmetaclass() inserts an extra class called 'NewBase' in the
        # inheritance tree: Entity -> NewBase -> object. But the initialization
        # should be executed only once for a given model class.

        # attrs will never be empty for classes declared in the standard way
        # (ie. with the `class` keyword). This is quite robust.
        if name == 'NewBase' and attrs == {}:
            return super_new(cls, name, bases, attrs)

        # Also ensure initialization is only performed for subclasses of Entity
        # (excluding Entity class itself).
        parents = [b for b in bases if isinstance(b, CommandMeta) and
                not (b.__name__ == 'NewBase' and b.__mro__ == (b, object))]
        if not parents:
            return super_new(cls, name, bases, attrs)
        module = attrs.pop('__module__')
        new_class = super_new(cls, name, bases, {'__module__': module,
            '__parameters__': tuple()})

        # Add all attributes to the new class.
        for attname, value in attrs.items():
            new_class.add_to_class(attname, value)

        attnames = map(lambda x: x.attname, new_class.__parameters__)
        new_class.add_to_class('tuple_type', collections.namedtuple(
            name + 'Tuple', attnames))
        new_class.__parameters__ = sorted(new_class.__parameters__,
            key=lambda x: x.creation_counter)
        new_class.__parameters__ = tuple(new_class.__parameters__)

        # Set qualified name
        new_class.qualified_name = module + '.' + name

        return new_class

    def add_to_class(cls, name, value):
        if hasattr(value, 'contribute_to_class'):
            value.contribute_to_class(cls, name)
        else:
            setattr(cls, name, value)



"""
Command-Query Responsibility Segregation framework (CQRS). Separates
destructive operations from non-destructive (read) operations; CQRS
is "the creation of two objects where there was previously one.
The separation occurs based upon whether the methods are a command
or a query." (Greg Young).

Example commands:
- Update the state of a relation in a relational database.
- Create a directory on a filesystem.

Example queries:
- Get a result set for a query on a relational database.
- Read the contents of a directory on a filesystem.
"""
import functools

from cqrs.command import Command
from cqrs.command import CommandAttribute
from cqrs.command import CommandParameter
from cqrs.command import CommandValidationError
from cqrs.command import CommandInspector
from cqrs.gateway import IGateway
from cqrs.exceptions import MissingCommandHandler
from cqrs.exceptions import UnprocessableCommandException
from cqrs.exceptions import CommandDeserializationError
from cqrs.exceptions import InvalidCommandParameter
from cqrs.exceptions import InvalidCommandParameters
from cqrs.exceptions import GatewayNotResponding
from cqrs.exceptions import GatewayRejectedCommand
from cqrs.exceptions import DuplicateCommand
from cqrs.exceptions import UnknownCommand
from cqrs.exceptions import InvalidRequestData
from cqrs.handler import CommandHandler


def get_version():
    return '1.0.1'






class CommandInspector(object):
    """
    Inspects and provides access to the internal state
    of :class:`~cqrs.Command` instances.
    """

    def getattr(self, command, attname):
        """Returns the command parameter."""
        return command._parameters[attname]

    def get_parameters(self, command):
        """
        Returns a :class:`tuple` of :class:`~cqrs.CommandParameter`
        instances specifying the parameters required by the command.
        """
        return command.__parameters__

    def as_dict(self, command):
        """
        Returns the parameters of a command as a Python dictionary,
        so that the output of :class:`CommandInspector.as_dict()`
        can be used to reinstantiate the object.
        """
        return {x.attname: self.getattr(command, x.attname)
            for x in command.__parameters__}

    def get_import_path(self, cls):
        """
        Returns the full dotted path of a class.
        """
        return cls.__module__ + '.' + cls.__name__

    def hash(self, command):
        """Implement the __hash__ method of a command."""
        return hash(command.as_tuple())

    def get_ident_attrs(self, command):
        return filter(lambda x: x.primary_key,
            self.get_parameters(command))


from libtng import cqrs


class TestCommand1(cqrs.Command):
    pass


@TestCommand1.handler
class TestCommandHandler1(cqrs.CommandHandler):
    pass


class TestCommand2(cqrs.Command):
    pass


@TestCommand2.handler
class TestCommandHandler2(cqrs.CommandHandler):
    pass






class MissingCommandHandler(ImportError):
    pass


class UnknownCommand(Exception):
    pass


class InvalidRequestData(Exception):
    pass


class UnprocessableCommandException(Exception):
    """
    Base class for exceptions raised when a command could
    not be processed.
    """
    http_status_code = None

    def __init__(self, http_status_code=None, *args, **kwargs):
        if http_status_code is not None:
            self.http_status_code = http_status_code

    def get_http_status_code(self):
        """
        Returns the HTTP status code associated with the exception.
        """
        return self.http_status_code or 500


class CommandDeserializationError(UnprocessableCommandException):
    """Raised when a command could not be deserialized by the gateway."""
    http_status_code = 422


class InvalidCommandParameter(UnprocessableCommandException):
    """
    Raised when a command could not be instatiated because of an invalid
    parameter. Mau be translated to a 422 UNPROCESSABLE ENTITY response.
    """
    http_status_code = 422


class InvalidCommandParameters(UnprocessableCommandException):
    """
    Raised when a command could not be instatiated because of invalid
    parameters. May be translated to a 422 UNPROCESSABLE ENTITY response.
    """
    http_status_code = 422

    def __init__(self, *args, **kwargs):
        self.invalid_parameters = kwargs.pop('parameters', [])
        UnprocessableCommandException.__init__(self, *args, **kwargs)


class GatewayNotResponding(UnprocessableCommandException):
    """
    Raised when a (remote) gateway is not responding. In a HTTP context
    this should translate to a 502 response; if the server does not want to
    disclose this information, use 503 instead.
    """
    http_status_code = 503


class GatewayRejectedCommand(UnprocessableCommandException):
    """
    Raised when a command was valid but the gateway rejected it. May be
    translated to a 422 HTTP response.
    """
    http_status_code = 422


class DuplicateCommand(UnprocessableCommandException):
    """
    Raised when an identical command is dispatched that should be
    unique within a certain timeframe; e.g. if the command
    ``RegisterSystemOperator('foo', 'mypassword')`` should be unique
    within an arbitrary timeframe. May be translated to a 409 CONFLICT
    response.
    """
    http_status_code = 409



#!/usr/bin/env python
from os.path import dirname
from os.path import abspath
from os.path import join
import os
import sys
import shutil
import glob

from distutils.core import setup
from distutils.sysconfig import get_python_lib

cwd = abspath(dirname(__file__))
PACKAGE_NAME = 'cqrs'
EXCLUDE_FROM_PACKAGES = []

# Warn if we are installing over top of an existing installation. This can
# cause issues where files that were deleted from a more recent version are
# still present in site-packages.
overlay_warning = False
if "install" in sys.argv:
    lib_paths = [get_python_lib()]
    if lib_paths[0].startswith("/usr/lib/"):
        # We have to try also with an explicit prefix of /usr/local in order to
        # catch Debian's custom user site-packages directory.
        lib_paths.append(get_python_lib(prefix="/usr/local"))
    for lib_path in lib_paths:
        existing_path = os.path.abspath(os.path.join(lib_path, PACKAGE_NAME))
        if os.path.exists(existing_path):
            # We note the need for the warning here, but present it after the
            # command is run, so it's more likely to be seen.
            overlay_warning = True
            break


def fullsplit(path, result=None):
    """
    Split a pathname into components (the opposite of os.path.join)
    in a platform-neutral way.
    """
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)


def is_package(package_name):
    for pkg in EXCLUDE_FROM_PACKAGES:
        if package_name.startswith(pkg):
            return False
    return True


# Compile the list of packages available, because distutils doesn't have
# an easy way to do this.
packages, package_data = [], {}

root_dir = os.path.dirname(__file__)
if root_dir != '':
    os.chdir(root_dir)
package_dir = PACKAGE_NAME

for dirpath, dirnames, filenames in os.walk(package_dir):
    # Ignore PEP 3147 cache dirs and those whose names start with '.'
    dirnames[:] = [d for d in dirnames if not d.startswith('.') and d != '__pycache__']
    parts = fullsplit(dirpath)
    package_name = '.'.join(parts)
    if '__init__.py' in filenames and is_package(package_name):
        packages.append(package_name)
    elif filenames:
        relative_path = []
        while '.'.join(parts) not in packages:
            relative_path.append(parts.pop())
        relative_path.reverse()
        path = os.path.join(*relative_path)
        package_files = package_data.setdefault('.'.join(parts), [])
        package_files.extend([os.path.join(path, f) for f in filenames])


# Dynamically calculate the version based on cqrs.VERSION.
version = __import__(PACKAGE_NAME).get_version()
data_dir = join(dirname(__file__), 'share')
data_files = [(d.replace(data_dir,'share/cqrs'), [join(d, f) for f in files]) for d,folders,files in os.walk(data_dir)]


setup(
    name='cqrs',
    version=version,
    url='http://www.tnginformationsystems.com/',
    author='Cochise Ruhulessin',
    author_email='cochise.ruhulessin@gmail.com',
    description="Command-Query Responsibility Separation (CQRS) framework",
    packages=packages,
    package_data=package_data,
    data_files=data_files,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Environment :: Other Environment',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Developers',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.4',
        'Topic :: Office/Business',

    ]
)
